
/* This software comprises contributed code made by Motorola, as a 
 * Contributor, to Network Simulator NS-2 software provided by the 
 * Regents of the University of California.
 * (Copyright; Regents of the University of California, 1994)
 * The contributed code was made as a result of a partnership between 
 * Motorola and INRIA Rhone-Alpes. 
 *
 * Copyright in the contributed code belongs to Motorola Inc. 2001
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ALL ADVERTISING MATERIALS MENTIONING FEATURES OR USE OF THIS SOFTWARE MUST 
 * DISPLAY AN ACKNOWLEDGEMENT TO THE COPYRIGHT OWNERS. 
 * ANY REDISTRIBUTION OF THIS SOFTWARE MUST CONTAIN THE ABOVE COPYRIGHT NOTICES, 
 * CONDITIONS AND DISCLAIMER.
 */

/* ############################################################################
 * This code was developed by Thierry Ernst (1998-2001)
 * MOTOROLA Labs Paris FRANCE - INRIA Rhone-Alpes Grenoble (PLANETE) FRANCE 
 * NS-2.1b6 enhancements for Wide-Area mobility simulations
 * ############################################################################
 */

// *************************************************************************
//                              ipv6.cc  
//      IPv6 Extensions  
//
// **************************************************************************
#include <iostream.h>
#include <address.h>
#include <node.h>
#include "agent.h"
#include "classifier.h"
#include "classifier-hash.h"
#include "classifier-src.h"
#include "ipv6.h"
#include "ipv6-routing.h"

// Check if I need the following include
extern "C" {
#include <stdarg.h>
};

#include <cmu-trace.h>
#include <mobilenode.h>
#include "packet.h"

#define PRINTADDR(a)     Address::instance().print_nodeaddr(a)

// **************************************************************
// Router Advertisement for Mobile IPv6 Base Stations
// **************************************************************
int hdr_rtads::offset_;
static class RTADSHeaderClass : public PacketHeaderClass {
public:
        RTADSHeaderClass() : PacketHeaderClass("PacketHeader/RTADS",
                                             sizeof(hdr_rtads)) {
            bind_offset(&hdr_rtads::offset_);
        }
} class_rtadshdr;

// **************************************************************
// Source Routing Extensions for Mobility Support in IPv6
// **************************************************************

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Source Routing Extensions.
// New Packet header to contain a list of IP addresses where the
// packet must transit
// Routing Header should only be inserted by the source node
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int hdr_rtext::offset_;
static class RtExtHeaderClass : public PacketHeaderClass {
public:
        RtExtHeaderClass() : PacketHeaderClass("PacketHeader/Routing",
					    sizeof(hdr_rtext)) {
		bind_offset(&hdr_rtext::offset_);
	}
} class_rtexthdr;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Source Routing Extensions.
// This class inserts a routing header in the packets
// The Routing Header contains a list of nodes where the packet
// must transit prior to reaching its final destination.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
static class SrcRoutingClass : public TclClass {
public:
        SrcRoutingClass() : TclClass("SrcRouting") {}
        TclObject* create(int, const char*const*) {
                return (new SrcRouting());
        }
} class_srcrouting;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Source Routing class definition 
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
SrcRouting::SrcRouting() : Connector(), node_(0) 
{
       bind("port_", &port_);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// If local node is the source of the packet, we have to insert
// the routing header - this also means we have an entry in the
// Binding Cache corresponding to destination.	
// XXX:  For now, only tested with one address in the routing header.
// XXX   Should also work with more than one (to be checked) 
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void SrcRouting::recv(Packet* p, Handler *h)
{
   Tcl& tcl = Tcl::instance();
   hdr_ip* iph = HDR_IP(p);
   hdr_cmn* cmh = HDR_CMN(p);
   hdr_rtext* rth = HDR_ROUTING(p);

   if ( iph->saddr() == node_->address() ) {

	// I need to insert an address in the header
	// Get destination address recorded in Binding Cache:
	// XXX: should rather access directly to the B Cache,
	// XXX  instead of calling OTcl.  Change this later. 
	int daddr = Address::instance().get_nodeaddr(iph->daddr());
	tcl.evalf("%s lookup-binding-cache %d", name_, daddr);
	int te = atoi(tcl.result());

	// Add the new transit address in the header
	ns_hop *hop = new ns_hop;
	hop->dst_ = iph->dst();
	hop->next_ = rth->nexthop_;  
	rth->nexthop_ = hop;
	rth->nb() ++;

	// Because Routing Header is processed after port_dmux 
	// for ease of use:
	iph->daddr() = te;
	iph->dport() = port_;

	// Size of routing header depends on number of addresses 
	// recorded in it:
	cmh->size() += IPv6_RTHDR_SIZE(rth->nb());

   } else {
	// If packet reaches here, this means this node is the 
	// destination of the packet as specified in the IP dst field.
	//  We may check this, but there is no reason to be wrong (...) 

	// I need to switch dst with address contained in header
	assert ( rth->nb() > 0 ); 
	rth->nb()--;
	iph->dst() = rth->nexthop_->dst_;
	rth->nexthop_ = rth->nexthop_->next_;
   } 
   // Pass packet to next object. 
   target_->recv(p,h);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int SrcRouting::command(int argc, const char*const* argv)
{
   if (argc == 3) {
      if (strcmp(argv[1], "node") == 0) {
        node_ = (Node*)TclObject::lookup(argv[2]);
	if (node_ == 0) {
           fprintf (stderr, "%s: %s lookup of %s failed\n", __FILE__, argv[1], 
argv[2]);
           return TCL_ERROR;
        }
        return TCL_OK;
      }
   }
   return (Connector::command(argc, argv));
}
 
// **************************************************************
// Classification by source address
// Adapted from classifier-hash.cc
// **************************************************************
static class SrcHashClassifierClass : public TclClass {
public:
        SrcHashClassifierClass() : TclClass("Classifier/Hash/Src") {}
        TclObject* create(int, const char*const*) {
                return new SrcHashClassifier;
        }

} class_hash_src_classifier;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int SrcHashClassifier::classify(Packet *p)
{
        int slot= lookup(p);
        if (slot >= 0 && slot <=maxslot_)
                return (slot);
        else if (default_ >= 0)
                return (default_);
        return -1;
} // HashClassifier::classify

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int SrcHashClassifier::command(int argc, const char*const* argv)
{
   if (argc == 4) {
        // $classifier install $src $node
        if (strcmp(argv[1], "install") == 0) {
           nsaddr_t src = atoi(argv[2]);
           NsObject *node = (NsObject*)TclObject::lookup(argv[3]);
           int slot = getnxt(node);
           install(slot, node);
           if (set_hash(src, 0, 0, slot) >= 0)
               return TCL_OK;
           else
               return TCL_ERROR;
           } // if
   }
   return(HashClassifier::command(argc, argv));
} // command


// *************************************************************************
// >----------------------------------------------------------------------<
// >                       Network Agent                                  <
// >----------------------------------------------------------------------<
// *************************************************************************

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
static void 
mac_callback (Packet * p, void *arg)
{
  // ((Network *) arg)->lost_link (p);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void 
Network::sendOutBCastPkt(Packet *p)
{
   hdr_cmn *cmh = HDR_CMN(p);
   cmh->next_hop_ = IP_BROADCAST; 
   cmh->addr_type_ = NS_AF_INET; 
   target_->recv(p, (Handler *)0);
}

Network::Network (): Agent (PT_MESSAGE), myaddr_(0), node_(0), port_dmux_(0), 
print_info_(0)
 
{
  bind ("print_info_", &print_info_);
}

// *************************************************************************
int 
Network::command (int argc, const char *const *argv)
{
  if (argc == 2)
    {
      if (strcmp (argv[1], "start-site") == 0)
	{
	  // Nothing for now 
	  return (TCL_OK);
	}
    }
  else if (argc == 3)
    {
      if (strcasecmp (argv[1], "addr") == 0) {
	 myaddr_ = Address::instance().str2addr(argv[2]);
	 return TCL_OK;
      }
      TclObject *obj;
      if ((obj = TclObject::lookup (argv[2])) == 0)
	{
	  fprintf (stderr, "%s: %s lookup of %s failed\n", __FILE__, argv[1],
		   argv[2]);
	  return TCL_ERROR;
	}
      if (strcasecmp (argv[1], "tracetarget") == 0)
	{
	  // tracetarget = (Trace *) obj;
	  return TCL_OK;
	}
      else if (strcasecmp (argv[1], "node") == 0) {
	      node_ = (MobileNode*) obj;
	      return TCL_OK;
      }
      else if (strcasecmp (argv[1], "port-dmux") == 0) {
	      port_dmux_ = (NsObject *) obj;
	      return TCL_OK;
      }
    }
  
  return (Agent::command (argc, argv));
}

// *************************************************************************
// >----------------------------------------------------------------------<
// >                    Network BS Agent                                  <
// >----------------------------------------------------------------------<
// *************************************************************************
void
NetworkBS::recv (Packet * p, Handler *)
{
  hdr_ip *iph = HDR_IP(p);
  hdr_cmn *cmh = HDR_CMN(p);
  int src = Address::instance().get_nodeaddr(iph->saddr());

  if (src == myaddr_) {
     // Packet I'm originating
     if (cmh->num_forwards() != 0) {
	// I received a packet that I sent.  
	// Probably  a routing loop.
	if (print_info_) cout << "\tNetworkBS - Routing Loop - drop\n";
	drop(p, DROP_RTR_ROUTE_LOOP);
	return;
     }
     else {
	// set direction of pkt to -1 , i.e downward
	cmh->direction() = hdr_cmn::DOWN;
	cmh->prev_hop_ = myaddr_;
	iph->ttl_ = IP_DEF_TTL;
		
        if (iph->daddr() == IP_BROADCAST) {
	   sendOutBCastPkt(p);
	}
	else {
	   if (print_info_) cout << "\tNetworkBS - HA forwarding to MN\n";
	   
	   cmh->addr_type_ = NS_AF_INET;
	   cmh->xmit_failure_ = mac_callback;
	   cmh->xmit_failure_data_ = this;
	   cmh->next_hop_ = Address::instance().get_nodeaddr(iph->daddr());
	   assert (!HDR_CMN (p)->xmit_failure_ ||
		HDR_CMN (p)->xmit_failure_ == mac_callback);
	   target_->recv(p, (Handler *)0);
	}
     }	
  } 
  else {
     	// Packet I receive
   	if (print_info_) {
		cout << "\tNetworkBS " << PRINTADDR(myaddr_);
		cout << " from " << PRINTADDR(src);
  	 }	

	if (--iph->ttl_ == 0) {
	   // Packet I'm forwarding...   
	   // Check the TTL.  If it is zero, then discard.
	   if (print_info_) cout << " - TTL expired - drop\n";
	   drop(p, DROP_RTR_TTL);
	   return;
	} 
	else  if ((u_int32_t) iph->daddr() == IP_BROADCAST) {
	   // Broadcast that I receive 
	   // hand it over to the port-demux
	   if (print_info_) cout << " - receiving BROADCAST\n";
	   port_dmux_->recv(p, (Handler*)0);
	   }
	else {
	   // BS forwarding to wireless (Mobile Node)
	   if (print_info_)
	     cout << " - forward to MN " << PRINTADDR(iph->daddr()) << "\n";
	   cmh->direction() = hdr_cmn::DOWN;
	   cmh->prev_hop_ = myaddr_;
	   cmh->next_hop_ = Address::instance().get_nodeaddr(iph->daddr());
	   target_->recv(p, (Handler *)0);
	}
  }
return;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
static class NetworkBSClass:public TclClass
{
public:
  NetworkBSClass ():TclClass ("Agent/Network/NetworkBS") {}
  TclObject *create (int, const char *const *) {
    return (new NetworkBS);
  }
} class_networkbs;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
NetworkBS::NetworkBS() : Network()
{
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int NetworkBS::command (int argc, const char *const *argv)
{
  return (Network::command (argc, argv));
}

// *************************************************************************
// >----------------------------------------------------------------------<
// >                    Network MN Agent                                  <
// >----------------------------------------------------------------------<
// *************************************************************************
static class NetworkMNClass:public TclClass
{
public:
  NetworkMNClass ():TclClass ("Agent/Network/NetworkMN") {}
  TclObject *create (int, const char *const *) {
    return (new NetworkMN);
  }
} class_networkmn;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
NetworkMN::NetworkMN() : Network(), mipagent_(0), decap_port_(-1)
{
	bind("off_ipinip_", &off_ipinip_);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void
NetworkMN::recv (Packet * p, Handler *)
{
  hdr_ip *iph = HDR_IP(p);
  hdr_cmn *cmh = HDR_CMN(p);
  int src = Address::instance().get_nodeaddr(iph->saddr());

  if (src == myaddr_) {
     // Packet I'm originating
     if (cmh->num_forwards() != 0) {
        // I received a packet that I sent.  
        // Probably  a routing loop.
	if (print_info_) cout << "\tNetworkMN - Routing Loop - drop\n";
        drop(p, DROP_RTR_ROUTE_LOOP);
        return;
     }
     else {
        // set direction of pkt to -1 , i.e downward
        cmh->direction() = hdr_cmn::DOWN;
        cmh->prev_hop_ = myaddr_;
        iph->ttl_ = IP_DEF_TTL;
                
        if ( iph->daddr() == IP_BROADCAST) {
           sendOutBCastPkt(p);
        }
        else {
	   // I am a Mobile IPv6 Node
	   // Forward DTG to my current base station
	   if (print_info_) 
		cout << "\tNetworkMN " << PRINTADDR(myaddr_)
		<< " from " << PRINTADDR(src)
		<< " forward to BS " << PRINTADDR(node_->base_stn()) << "\n";
           
           cmh->addr_type_ = NS_AF_INET;
           cmh->xmit_failure_ = mac_callback;
           cmh->xmit_failure_data_ = this;
	   cmh->next_hop_ = node_->base_stn();
           assert (!HDR_CMN (p)->xmit_failure_ ||
                HDR_CMN (p)->xmit_failure_ == mac_callback);
           target_->recv(p, (Handler *)0);
        }
     }  
  } 
  else {
        // Packet I receive
   	if (print_info_) {
		cout << "\tNetworkMN " << PRINTADDR(myaddr_)
		<< " from " << PRINTADDR(src) << " to " 
		<< PRINTADDR(iph->daddr());
   	} 

        if (--iph->ttl_ == 0) {
           // Packet I'm forwarding...   
           // Check the TTL.  If it is zero, then discard.
	   if (print_info_) cout << " - TTL expired - drop\n";
           drop(p, DROP_RTR_TTL);
           return;
        } 
        else  if ((u_int32_t) iph->daddr() == IP_BROADCAST) {
           // Broadcast that I receive 
	   // (Likely a Router Advertisement)
           // hand it over to the port-demux
	   if (print_info_) cout << " - receiving BROADCAST\n";
           port_dmux_->recv(p, (Handler*)0);
           }
        else {
	   // DTG for myself
	   // Probably a packet with destination = COA (with routing header 
	   // or encapsulation) otherwise may have not reach here 
	   if (print_info_) cout << " - receiving DTG " << cmh->ptype_ <<"\n";

	   // If it comes encapsulated, we may want to send BU to 
	   // src addr in the encapsulated header.  We add the 
	   // source in the Binding List.
	   // XXX: not very cool: we check inside header before
	   // packet is effectively decapsulated ... 
	   if ( iph->dport() == decap_port_ ) {
	      hdr_ipinip **ppinhdr = (hdr_ipinip **)p->access(off_ipinip_);
	      hdr_ip *pinhdr = &(*ppinhdr)->hdr_;
	      mipagent_->add_bulist(pinhdr->saddr(), BU_CN, ON);
	   } else {
	      mipagent_->add_bulist(iph->saddr(), BU_CN, ON);
	   } 
	   port_dmux_->recv(p, (Handler*)0);
        }
  }
return;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int NetworkMN::command (int argc, const char *const *argv)
{
  if ( argc == 3 ) {
     if (strcmp(argv[1], "mip-agent") == 0) {
        mipagent_ = (MNAgent*)TclObject::lookup(argv[2]); 
        return TCL_OK;
     }
     if (strcmp(argv[1], "decap-port") == 0) {
        decap_port_ = atoi(argv[2]); 
        return TCL_OK;
     }
  }
  return (Network::command (argc, argv));
  
}




